@@ -23,7 +23,6 @@
 #include <grub/mm.h>
 #include <grub/usb.h>
 #include <grub/usbserial.h>
-#include <grub/i18n.h>
 
 GRUB_MOD_LICENSE ("GPLv3+");
 
@@ -82,7 +81,6 @@ real_config (struct grub_serial_port *port)
   };
   const grub_uint16_t stop_bits[] = {
     [GRUB_SERIAL_STOP_BITS_1] = 0x0000,
-    [GRUB_SERIAL_STOP_BITS_1_5] = 0x0800,
     [GRUB_SERIAL_STOP_BITS_2] = 0x1000,
   };
 
@@ -139,24 +137,19 @@ ftdi_hw_configure (struct grub_serial_port *port,
 
   divisor = get_divisor (config->speed);
   if (divisor == 0)
-    return grub_error (GRUB_ERR_BAD_ARGUMENT,
-		       N_("unsupported serial port speed"));
+    return grub_error (GRUB_ERR_BAD_ARGUMENT, "bad speed");
 
   if (config->parity != GRUB_SERIAL_PARITY_NONE
       && config->parity != GRUB_SERIAL_PARITY_ODD
       && config->parity != GRUB_SERIAL_PARITY_EVEN)
-    return grub_error (GRUB_ERR_BAD_ARGUMENT,
-		       N_("unsupported serial port parity"));
+    return grub_error (GRUB_ERR_BAD_ARGUMENT, "unsupported parity");
 
   if (config->stop_bits != GRUB_SERIAL_STOP_BITS_1
-      && config->stop_bits != GRUB_SERIAL_STOP_BITS_1_5
       && config->stop_bits != GRUB_SERIAL_STOP_BITS_2)
-    return grub_error (GRUB_ERR_BAD_ARGUMENT,
-		       N_("unsupported serial port stop bits number"));
+    return grub_error (GRUB_ERR_BAD_ARGUMENT, "unsupported stop bits");
 
   if (config->word_len < 5 || config->word_len > 8)
-    return grub_error (GRUB_ERR_BAD_ARGUMENT,
-		       N_("unsupported serial port word length"));
+    return grub_error (GRUB_ERR_BAD_ARGUMENT, "unsupported word length");
 
   port->config = *config;
   port->configured = 0;
