@@ -31,7 +31,6 @@
 #include <grub/extcmd.h>
 #include <grub/datetime.h>
 #include <grub/i18n.h>
-#include <grub/net.h>
 
 GRUB_MOD_LICENSE ("GPLv3+");
 
@@ -61,22 +60,6 @@ grub_ls_list_devices (int longlist)
 
   grub_device_iterate (grub_ls_print_devices);
   grub_xputs ("\n");
-
-#if 0
-  {
-    grub_net_app_level_t proto;
-    int first = 1;
-    FOR_NET_APP_LEVEL (proto)
-    {
-      if (first)
-	grub_puts_ (N_ ("Network protocols:"));
-      first = 0;
-      grub_printf ("%s ", proto->name);
-    }
-    grub_xputs ("\n");
-  }
-#endif
-
   grub_refresh ();
 
   return 0;
@@ -138,7 +121,7 @@ grub_ls_list_files (char *dirname, int longlist, int all, int human)
 	  else
 	    {
 	      grub_uint64_t fsize = file->size * 100ULL;
-	      grub_uint64_t fsz = file->size;
+	      int fsz = file->size;
 	      int units = 0;
 	      char buf[20];
 
@@ -151,12 +134,11 @@ grub_ls_list_files (char *dirname, int longlist, int all, int human)
 
 	      if (units)
 		{
-		  grub_uint64_t whole, fraction;
+		  grub_uint32_t whole, fraction;
 
 		  whole = grub_divmod64 (fsize, 100, &fraction);
 		  grub_snprintf (buf, sizeof (buf),
-				 "%" PRIuGRUB_UINT64_T
-				 ".%02" PRIuGRUB_UINT64_T "%c", whole, fraction,
+				 "%u.%02u%c", whole, fraction,
 				 grub_human_sizes[units]);
 		  grub_printf ("%-12s", buf);
 		}
@@ -168,7 +150,7 @@ grub_ls_list_files (char *dirname, int longlist, int all, int human)
 	  grub_free (pathname);
 	}
       else
-	grub_printf ("%-12s", _("DIR"));
+	grub_printf ("%-12s", "DIR");
 
       if (info->mtimeset)
 	{
