@@ -121,9 +121,11 @@ grub_efiemu_register_prepare_hook (grub_err_t (*hook) (void *data),
 				   void *data)
 {
   struct grub_efiemu_prepare_hook *nhook;
+  if (! hook)
+    return grub_error (GRUB_ERR_BAD_ARGUMENT, "you must supply the hook");
   nhook = (struct grub_efiemu_prepare_hook *) grub_malloc (sizeof (*nhook));
   if (! nhook)
-    return grub_errno;
+    return grub_error (GRUB_ERR_OUT_OF_MEMORY, "couldn't prepare hook");
   nhook->hook = hook;
   nhook->unload = unload;
   nhook->data = data;
@@ -144,13 +146,15 @@ grub_efiemu_register_configuration_table (grub_efi_guid_t guid,
   struct grub_efiemu_configuration_table *tbl;
   grub_err_t err;
 
- err = grub_efiemu_unregister_configuration_table (guid);
-  if (err)
+ if (! get_table && ! data)
+    return grub_error (GRUB_ERR_BAD_ARGUMENT,
+		       "you must set at least get_table or data");
+  if ((err = grub_efiemu_unregister_configuration_table (guid)))
     return err;
 
   tbl = (struct grub_efiemu_configuration_table *) grub_malloc (sizeof (*tbl));
   if (! tbl)
-    return grub_errno;
+    return grub_error (GRUB_ERR_OUT_OF_MEMORY, "couldn't register table");
 
   tbl->guid = guid;
   tbl->get_table = get_table;
@@ -196,12 +200,12 @@ grub_efiemu_load_file (const char *filename)
     {
       grub_file_close (file);
       grub_efiemu_unload ();
-      return grub_errno;
+      return grub_error (grub_errno, "couldn't init memory management");
     }
 
   grub_dprintf ("efiemu", "mm initialized\n");
 
-  err = grub_efiemu_loadcore_init (file, filename);
+  err = grub_efiemu_loadcore_init (file);
   if (err)
     {
       grub_file_close (file);
@@ -222,7 +226,7 @@ grub_efiemu_autocore (void)
 {
   const char *prefix;
   char *filename;
-  const char *suffix;
+  char *suffix;
   grub_err_t err;
 
   if (grub_efiemu_sizeof_uintn_t () != 0)
@@ -232,14 +236,16 @@ grub_efiemu_autocore (void)
 
   if (! prefix)
     return grub_error (GRUB_ERR_FILE_NOT_FOUND,
-		       N_("variable `%s' isn't set"), "prefix");
+		       "couldn't find efiemu core because prefix "
+		       "isn't set");
 
   suffix = grub_efiemu_get_default_core_name ();
 
-  filename = grub_xasprintf ("%s/" GRUB_TARGET_CPU "-" GRUB_PLATFORM "/%s",
-			     prefix, suffix);
+  filename = grub_xasprintf ("%s/%s", prefix, suffix);
   if (! filename)
-    return grub_errno;
+    return grub_error (GRUB_ERR_OUT_OF_MEMORY,
+		       "couldn't allocate temporary space");
+
 
   err = grub_efiemu_load_file (filename);
   grub_free (filename);
@@ -290,7 +296,7 @@ grub_cmd_efiemu_load (grub_command_t cmd __attribute__ ((unused)),
   grub_efiemu_unload ();
 
   if (argc != 1)
-    return grub_error (GRUB_ERR_BAD_ARGUMENT, N_("filename expected"));
+    return grub_error (GRUB_ERR_BAD_ARGUMENT, "filename required");
 
   err = grub_efiemu_load_file (args[0]);
   if (err)
