@@ -62,17 +62,12 @@ grub_efiemu_add_to_mmap (grub_uint64_t start, grub_uint64_t size,
   /* Extend map if necessary*/
   if (mmap_num >= mmap_reserved_size)
     {
-      void *old;
-      mmap_reserved_size = 2 * (mmap_reserved_size + 1);
-      old = efiemu_mmap;
       efiemu_mmap = (grub_efi_memory_descriptor_t *)
-	grub_realloc (efiemu_mmap, mmap_reserved_size
+	grub_realloc (efiemu_mmap, (++mmap_reserved_size)
 		      * sizeof (grub_efi_memory_descriptor_t));
       if (!efiemu_mmap)
-	{
-	  grub_free (old);
-	  return grub_errno;
-	}
+	return grub_error (GRUB_ERR_OUT_OF_MEMORY,
+			   "not enough space for memory map");
     }
 
   /* Fill slot*/
@@ -181,7 +176,8 @@ efiemu_alloc_requests (void)
   /* Allocate the whole memory in one block */
   resident_memory = grub_memalign (GRUB_EFIEMU_PAGESIZE, total_alloc);
   if (!resident_memory)
-    return grub_errno;
+    return grub_error (GRUB_ERR_OUT_OF_MEMORY,
+		       "couldn't allocate resident memory");
 
   /* Split the memory into blocks by type */
   curptr = resident_memory;
@@ -205,10 +201,10 @@ efiemu_alloc_requests (void)
 	- (requested_memory[reqorder[i]] % GRUB_EFIEMU_PAGESIZE);
       if (align_overhead == GRUB_EFIEMU_PAGESIZE)
 	align_overhead = 0;
-      curptr = ((grub_uint8_t *) curptr) + align_overhead;
+      curptr = ((grub_uint8_t *)curptr) + align_overhead;
 
       /* Add the region to memory map */
-      grub_efiemu_add_to_mmap ((grub_addr_t) typestart,
+      grub_efiemu_add_to_mmap (PTR_TO_UINT64 (typestart),
 			       curptr - typestart, reqorder[i]);
     }
 
@@ -408,8 +404,7 @@ grub_efiemu_mmap_fill (void)
 					  GRUB_EFI_ACPI_MEMORY_NVS);
 
 	default:
-	  grub_dprintf ("efiemu",
-			"Unknown memory type %d. Assuming unusable\n", type);
+	  grub_printf ("Unknown memory type %d. Assuming unusable\n", type);
 	case GRUB_MEMORY_RESERVED:
 	  return grub_efiemu_add_to_mmap (addr, size,
 					  GRUB_EFI_UNUSABLE_MEMORY);
@@ -544,7 +539,8 @@ grub_efiemu_mmap_sort_and_uniq (void)
     {
       grub_free (result);
       grub_free (scanline_events);
-      return grub_errno;
+      return grub_error (GRUB_ERR_OUT_OF_MEMORY,
+			 "couldn't allocate space for new memory map");
     }
 
   /* Register scanline events */
@@ -645,14 +641,12 @@ grub_efiemu_mm_do_alloc (void)
   if (!efiemu_mmap)
     {
       grub_efiemu_unload ();
-      return grub_errno;
+      return grub_error (GRUB_ERR_OUT_OF_MEMORY, "couldn't initialize mmap");
     }
 
-  err = efiemu_alloc_requests ();
-  if (err)
+  if ((err = efiemu_alloc_requests ()))
     return err;
-  err = grub_efiemu_mmap_fill ();
-  if (err)
+  if ((err = grub_efiemu_mmap_fill ()))
     return err;
   return grub_efiemu_mmap_sort_and_uniq ();
 }
