@@ -53,13 +53,13 @@ static struct grub_partition_map grub_sun_pc_partition_map;
 
 /* Verify checksum (true=ok).  */
 static int
-grub_sun_is_valid (grub_uint16_t *label)
+grub_sun_is_valid (struct grub_sun_pc_block *label)
 {
   grub_uint16_t *pos;
   grub_uint16_t sum = 0;
 
-  for (pos = label;
-       pos < (label + sizeof (struct grub_sun_pc_block) / 2);
+  for (pos = (grub_uint16_t *) label;
+       pos < (grub_uint16_t *) (label + 1);
        pos++)
     sum ^= *pos;
 
@@ -72,11 +72,7 @@ sun_pc_partition_map_iterate (grub_disk_t disk,
 					   const grub_partition_t partition))
 {
   grub_partition_t p;
-  union
-  {
-    struct grub_sun_pc_block sun;
-    grub_uint16_t raw[0];
-  } block;
+  struct grub_sun_pc_block block;
   int partnum;
   grub_err_t err;
 
@@ -92,14 +88,14 @@ sun_pc_partition_map_iterate (grub_disk_t disk,
       return err;
     }
   
-  if (GRUB_PARTMAP_SUN_PC_MAGIC != grub_le_to_cpu16 (block.sun.magic))
+  if (GRUB_PARTMAP_SUN_PC_MAGIC != grub_le_to_cpu16 (block.magic))
     {
       grub_free (p);
       return grub_error (GRUB_ERR_BAD_PART_TABLE, 
 			 "not a sun_pc partition table");
     }
 
-  if (! grub_sun_is_valid (block.raw))
+  if (! grub_sun_is_valid (&block))
     {
       grub_free (p);
       return grub_error (GRUB_ERR_BAD_PART_TABLE, "invalid checksum");
@@ -111,12 +107,11 @@ sun_pc_partition_map_iterate (grub_disk_t disk,
     {
       struct grub_sun_pc_partition_descriptor *desc;
 
-      if (block.sun.partitions[partnum].id == 0
-	  || block.sun.partitions[partnum].id
-	  == GRUB_PARTMAP_SUN_PC_WHOLE_DISK_ID)
+      if (block.partitions[partnum].id == 0
+	  || block.partitions[partnum].id == GRUB_PARTMAP_SUN_PC_WHOLE_DISK_ID)
 	continue;
 
-      desc = &block.sun.partitions[partnum];
+      desc = &block.partitions[partnum];
       p->start = grub_le_to_cpu32 (desc->start_sector);
       p->len = grub_le_to_cpu32 (desc->num_sectors);
       p->number = partnum;
