@@ -18,7 +18,6 @@
 
 #define grub_video_render_target grub_video_fbrender_target
 
-#if !defined (TEST) && !defined(GENINIT)
 #include <grub/err.h>
 #include <grub/types.h>
 #include <grub/dl.h>
@@ -28,21 +27,9 @@
 #include <grub/video_fb.h>
 #include <grub/pci.h>
 #include <grub/vga.h>
-#include <grub/cache.h>
-#else
-typedef unsigned char grub_uint8_t;
-typedef unsigned short grub_uint16_t;
-typedef unsigned int grub_uint32_t;
-typedef int grub_err_t;
-#include <grub/vgaregs.h>
-#include <stdio.h>
-#define ARRAY_SIZE(array) (sizeof (array) / sizeof (array[0]))
-#endif
 
 #include "sm712_init.c"
 
-#pragma GCC diagnostic ignored "-Wcast-align"
-
 #define GRUB_SM712_TOTAL_MEMORY_SPACE  0x700400
 #define GRUB_SM712_REG_BASE 0x700000
 #define GRUB_SM712_PCIID 0x0712126f
@@ -204,20 +191,16 @@ static struct
 
 static struct
 {
-#if !defined (TEST) && !defined(GENINIT)
   struct grub_video_mode_info mode_info;
-#endif
+  struct grub_video_render_target *render_target;
 
-  volatile grub_uint8_t *ptr;
-  grub_uint8_t *cached_ptr;
+  grub_uint8_t *ptr;
   int mapped;
   grub_uint32_t base;
-#if !defined (TEST) && !defined(GENINIT)
   grub_pci_device_t dev;
-#endif
 } framebuffer;
 
-#if !defined (TEST) && !defined(GENINIT)
+#ifndef TEST
 static grub_err_t
 grub_video_sm712_video_init (void)
 {
@@ -231,12 +214,9 @@ static grub_err_t
 grub_video_sm712_video_fini (void)
 {
   if (framebuffer.mapped)
-    {
-      grub_pci_device_unmap_range (framebuffer.dev, framebuffer.ptr,
-				   GRUB_SM712_TOTAL_MEMORY_SPACE);
-      grub_pci_device_unmap_range (framebuffer.dev, framebuffer.cached_ptr,
-				   GRUB_SM712_TOTAL_MEMORY_SPACE);
-    }
+    grub_pci_device_unmap_range (framebuffer.dev, framebuffer.ptr,
+				 GRUB_SM712_TOTAL_MEMORY_SPACE);
+
   return grub_video_fb_fini ();
 }
 #endif
@@ -246,10 +226,6 @@ grub_sm712_write_reg (grub_uint8_t val, grub_uint16_t addr)
 {
 #ifdef TEST
   printf ("  {1, 0x%x, 0x%x},\n", addr, val);
-#elif defined (GENINIT)
-  printf (" .byte 0x%02x, 0x%02x\n", (addr - 0x3c0), val);
-  if ((addr - 0x3c0) & ~0x7f)
-    printf ("FAIL\n");
 #else
    *(volatile grub_uint8_t *) (framebuffer.ptr + GRUB_SM712_REG_BASE
 			       + addr) = val;
@@ -261,10 +237,6 @@ grub_sm712_read_reg (grub_uint16_t addr)
 {
 #ifdef TEST
   printf ("  {-1, 0x%x, 0x5},\n", addr);
-#elif defined (GENINIT)
-  if ((addr - 0x3c0) & ~0x7f)
-    printf ("FAIL\n");
-  printf ("  .byte 0x%04x, 0x00\n", (addr - 0x3c0) | 0x80);
 #else
   return *(volatile grub_uint8_t *) (framebuffer.ptr + GRUB_SM712_REG_BASE
 				     + addr);
@@ -364,12 +336,12 @@ static grub_err_t
 grub_video_sm712_setup (unsigned int width, unsigned int height,
 			unsigned int mode_type, unsigned int mode_mask __attribute__ ((unused)))
 {
-  unsigned i;
-#if !defined (TEST) && !defined(GENINIT)
   int depth;
   grub_err_t err;
   int found = 0;
+  unsigned i;
 
+#ifndef TEST
   auto int NESTED_FUNC_ATTR find_card (grub_pci_device_t dev, grub_pci_id_t pciid __attribute__ ((unused)));
   int NESTED_FUNC_ATTR find_card (grub_pci_device_t dev, grub_pci_id_t pciid __attribute__ ((unused)))
     {
@@ -404,12 +376,11 @@ grub_video_sm712_setup (unsigned int width, unsigned int height,
   grub_pci_iterate (find_card);
   if (!found)
     return grub_error (GRUB_ERR_IO, "Couldn't find graphics card");
+#endif
   /* Fill mode info details.  */
   framebuffer.mode_info.width = 1024;
   framebuffer.mode_info.height = 600;
-  framebuffer.mode_info.mode_type = (GRUB_VIDEO_MODE_TYPE_RGB
-				     | GRUB_VIDEO_MODE_TYPE_DOUBLE_BUFFERED
-				     | GRUB_VIDEO_MODE_TYPE_UPDATING_SWAP);
+  framebuffer.mode_info.mode_type = GRUB_VIDEO_MODE_TYPE_RGB;
   framebuffer.mode_info.bpp = 16;
   framebuffer.mode_info.bytes_per_pixel = 2;
   framebuffer.mode_info.pitch = 1024 * 2;
@@ -422,11 +393,12 @@ grub_video_sm712_setup (unsigned int width, unsigned int height,
   framebuffer.mode_info.blue_field_pos = 0;
   framebuffer.mode_info.reserved_mask_size = 0;
   framebuffer.mode_info.reserved_field_pos = 0;
+#ifndef TEST
   framebuffer.mode_info.blit_format
     = grub_video_get_blit_format (&framebuffer.mode_info);
 #endif
 
-#if !defined (TEST) && !defined(GENINIT)
+#ifndef TEST
   if (found && framebuffer.base == 0)
     {
       grub_pci_address_t addr;
@@ -447,27 +419,23 @@ grub_video_sm712_setup (unsigned int width, unsigned int height,
 #endif
 
   /* We can safely discard volatile attribute.  */
-#if !defined (TEST) && !defined(GENINIT)
+#ifndef TEST
   framebuffer.ptr
-    = grub_pci_device_map_range (framebuffer.dev,
-				 framebuffer.base,
-				 GRUB_SM712_TOTAL_MEMORY_SPACE);
-  framebuffer.cached_ptr
-    = grub_pci_device_map_range_cached (framebuffer.dev,
-					framebuffer.base,
-					GRUB_SM712_TOTAL_MEMORY_SPACE);
+    = (void *) grub_pci_device_map_range (framebuffer.dev,
+					  framebuffer.base,
+					  GRUB_SM712_TOTAL_MEMORY_SPACE);
 #endif
   framebuffer.mapped = 1;
 
   /* Initialise SM712.  */
-#if !defined (TEST) && !defined(GENINIT)
+#ifndef TEST
   /* FIXME */
   grub_vga_sr_write (0x11, 0x18);
 #endif
 
-#if !defined (TEST) && !defined(GENINIT)
+#ifndef TEST
   /* Prevent garbage from appearing on the screen.  */
-  grub_memset ((void *) framebuffer.cached_ptr, 0, 
+  grub_memset (framebuffer.ptr, 0, 
 	       framebuffer.mode_info.height * framebuffer.mode_info.pitch);
 #endif
 
@@ -707,7 +675,7 @@ grub_video_sm712_setup (unsigned int width, unsigned int height,
 			| GRUB_VGA_IO_MISC_COLOR,
 			GRUB_VGA_IO_MISC_WRITE);
 
-#if !defined (TEST) && !defined(GENINIT)
+#ifndef TEST
   /* Undocumented? */
   *(volatile grub_uint32_t *) ((char *) framebuffer.ptr + 0x40c00c) = 0;
   *(volatile grub_uint32_t *) ((char *) framebuffer.ptr + 0x40c040) = 0;
@@ -717,36 +685,45 @@ grub_video_sm712_setup (unsigned int width, unsigned int height,
 
   (void) grub_sm712_sr_read (0x16);
 
-#if !defined (TEST) && !defined(GENINIT)
-  err = grub_video_fb_setup (mode_type, mode_mask,
-			     &framebuffer.mode_info,
-			     framebuffer.cached_ptr, NULL, NULL);
+#ifndef TEST
+  err = grub_video_fb_create_render_target_from_pointer (&framebuffer
+							 .render_target,
+							 &framebuffer.mode_info,
+							 framebuffer.ptr);
+
+  if (err)
+    return err;
+
+  err = grub_video_fb_set_active_render_target (framebuffer.render_target);
+  
   if (err)
     return err;
 
   /* Copy default palette to initialize emulated palette.  */
   err = grub_video_fb_set_palette (0, GRUB_VIDEO_FBSTD_NUMCOLORS,
 				   grub_video_fbstd_colors);
-  return err;
-#else
-  return 0;
 #endif
+  return err;
 }
 
-#if !defined (TEST) && !defined(GENINIT)
+#ifndef TEST
 
 static grub_err_t
 grub_video_sm712_swap_buffers (void)
 {
-  grub_size_t s;
-  s = (framebuffer.mode_info.height
-       * framebuffer.mode_info.pitch
-       * framebuffer.mode_info.bytes_per_pixel);
-  grub_video_fb_swap_buffers ();
-  grub_arch_sync_dma_caches (framebuffer.cached_ptr, s);
+  /* TODO: Implement buffer swapping.  */
   return GRUB_ERR_NONE;
 }
 
+static grub_err_t
+grub_video_sm712_set_active_render_target (struct grub_video_render_target *target)
+{
+  if (target == GRUB_VIDEO_RENDER_TARGET_DISPLAY)
+      target = framebuffer.render_target;
+
+  return grub_video_fb_set_active_render_target (target);
+}
+
 static grub_err_t
 grub_video_sm712_get_info_and_fini (struct grub_video_mode_info *mode_info,
 				    void **framebuf)
@@ -786,26 +763,18 @@ static struct grub_video_adapter grub_video_sm712_adapter =
     .swap_buffers = grub_video_sm712_swap_buffers,
     .create_render_target = grub_video_fb_create_render_target,
     .delete_render_target = grub_video_fb_delete_render_target,
-    .set_active_render_target = grub_video_fb_set_active_render_target,
+    .set_active_render_target = grub_video_sm712_set_active_render_target,
     .get_active_render_target = grub_video_fb_get_active_render_target,
 
     .next = 0
   };
 
-#ifdef GRUB_MACHINE_MIPS_LOONGSON
-void grub_video_sm712_init (void)
-#else
 GRUB_MOD_INIT(video_sm712)
-#endif
 {
   grub_video_register (&grub_video_sm712_adapter);
 }
 
-#ifdef GRUB_MACHINE_MIPS_LOONGSON
-void grub_video_sm712_fini (void)
-#else
 GRUB_MOD_FINI(video_sm712)
-#endif
 {
   grub_video_unregister (&grub_video_sm712_adapter);
 }
